#    (c) Copyright 2019 Brocade, a Broadcom Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

YANG = "application/yang-data+json"
ACCEPT = "Accept"
CONTENT_TYPE = "Content-Type"
AUTHORIZATION = "Authorization"
USER_AGENT = "User-Agent"
ZONE_DRIVER = "OpenStack Zone Driver"
LOGIN = "/rest/login"
LOGOUT = "/rest/logout"
NAME_SERVER = "/rest/running/brocade-name-server"
ZONING = "/rest/running/zoning"
DEFINED_CFG = "/defined-configuration"
EFFECTIVE_CFG = "/effective-configuration"
GET_SWITCH = "/rest/running/switch/fibrechannel-switch"
GET_NAMESERVER = NAME_SERVER + "/fibrechannel-name-server"
GET_DEFINED_ZONE_CFG = ZONING + DEFINED_CFG
GET_ACTIVE_ZONE_CFG = ZONING + EFFECTIVE_CFG
GET_CHECKSUM = ZONING + EFFECTIVE_CFG + "/checksum"
POST_ZONE = ZONING + DEFINED_CFG + "/zone/zone-name/"
POST_CFG = ZONING + DEFINED_CFG + "/cfg/cfg-name/"
PATCH_CFG = ZONING + DEFINED_CFG + "/cfg/cfg-name/"
PATCH_CFG_SAVE = ZONING + EFFECTIVE_CFG + "/cfg-action/1"
PATCH_CFG_DISABLE = ZONING + EFFECTIVE_CFG + "/cfg-action/2"
PATCH_CFG_ENABLE = ZONING + EFFECTIVE_CFG + "/cfg-name/"
DELETE_ZONE = POST_ZONE
DELETE_CFG = POST_CFG
RESPONSE = "Response"
SWITCH = "fibrechannel-switch"
FIRMWARE_VERSION = "firmware-version"
FC_NAME_SERVER = "fibrechannel-name-server"
PORT_NAME = "port-name"
DEFINED_CFG = "defined-configuration"
CFG = "cfg"
CFG_NAME = "cfg-name"
MEMBER_ZONE = "member-zone"
ZONE_NAME = "zone-name"
ZONE = "zone"
MEMBER_ENTRY = "member-entry"
ENTRY_NAME = "entry-name"
ALIAS = "alias"
ALIAS_ENTRY_NAME = "alias-entry-name"
EFFECTIVE_CFG = "effective-configuration"
CHECKSUM = "checksum"
ENABLED_ZONE = "enabled-zone"



#! /usr/bin/env python
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Generate list of Cinder drivers"""

import argparse
import operator
import os
import json
import textwrap

from cinder.interface import util
from cinder import objects


# Object loading can cause issues loading drivers, force it up front
objects.register_all()


parser = argparse.ArgumentParser(prog="generate_driver_list")

parser.add_argument("--format", default='str', choices=['str', 'dict'],
                    help="Output format type")

# Keep backwards compatibilty with the gate-docs test
# The tests pass ['docs'] on the cmdln, but it's never been used.
parser.add_argument("output_list", default=None, nargs='?')

CI_WIKI_ROOT = "https://wiki.openstack.org/wiki/ThirdPartySystems/"


class Output(object):

    def __init__(self, base_dir, output_list):
        # At this point we don't care what was passed in, just a trigger
        # to write this out to the doc tree for now
        self.driver_file = None
        if output_list:
            self.driver_file = open(
                '%s/doc/source/drivers.rst' % base_dir, 'w+')
            self.driver_file.write('===================\n')
            self.driver_file.write('Available Drivers\n')
            self.driver_file.write('===================\n\n')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.driver_file:
            self.driver_file.close()

    def write(self, text):
        if self.driver_file:
            self.driver_file.write('%s\n' % text)
        else:
            print(text)


def format_description(desc, output):
    desc = desc or '<None>'
    lines = desc.rstrip('\n').split('\n')
    output.write('* Description: %s' % lines[0])
    output.write('')
    output.write(textwrap.dedent('\n'.join(lines[1:])))


def format_options(driver_options, output):
    if driver_options and len(driver_options) > 0:

        output.write('* Driver Configuration Options:')
        output.write('')
        output.write('.. list-table:: **Driver configuration options**')
        output.write('   :header-rows: 1')
        output.write('   :widths: 14 30')
        output.write('')
        output.write('   * - Name = Default Value')
        output.write('     - (Type) Description')
        sorted_options = sorted(driver_options,
                                key=operator.attrgetter('name'))
        for opt in sorted_options:
            output.write('   * - %s = %s' %
                         (opt.name, opt.default))
            output.write('     - (%s) %s' % (opt.type, opt.help))
        output.write('')

def filter_drivers(drivers):
    '''This filters all of the drivers into separate lists.'''

    supported_drivers = []
    unsupported_drivers = []

    for driver in drivers:
        if not driver.supported:
            unsupported_drivers.append(driver)
        else:
            supported_drivers.append(driver)

    return supported_drivers, unsupported_drivers


def print_drivers(drivers, config_name, output, section_char='-',
                  display_unsupported=True):
    for driver in sorted(drivers, key=lambda x: x.class_name):
        driver_name = driver.class_name
        if not driver.supported and display_unsupported:
            driver_name += " (unsupported)"
        output.write(driver_name)
        output.write(section_char * len(driver_name))
        if driver.version:
            output.write('* Version: %s' % driver.version)
        output.write('* %s=%s' % (config_name, driver.class_fqn))
        if driver.ci_wiki_name and 'Cinder_Jenkins' not in driver.ci_wiki_name:
            output.write('* CI info: %s%s' % (CI_WIKI_ROOT,
                                              driver.ci_wiki_name))
        format_options(driver.driver_options, output)
        format_description(driver.desc, output)
        output.write('')
    output.write('')


def output_str(cinder_root, args):
    with Output(cinder_root, args.output_list) as output:
        output.write('Volume Drivers')
        output.write('==============')
        supported_drivers, unsupported_drivers = filter_drivers(
            util.get_volume_drivers())

        output.write('Supported Drivers')
        output.write('-----------------')
        output.write('')
        print_drivers(supported_drivers, 'volume_driver', output, '~')

        output.write('Unsupported Drivers')
        output.write('-------------------')
        output.write('')
        print_drivers(unsupported_drivers, 'volume_driver', output, '~')

        output.write('Backup Drivers')
        output.write('==============')
        print_drivers(util.get_backup_drivers(), 'backup_driver', output)

        output.write('FC Zone Manager Drivers')
        output.write('=======================')
        print_drivers(util.get_fczm_drivers(), 'zone_driver', output)


def collect_driver_info(driver):
    """Build the dictionary that describes this driver."""

    info = {'name': driver.class_name,
            'version': driver.version,
            'fqn': driver.class_fqn,
            'description': driver.desc,
            'ci_wiki_name': driver.ci_wiki_name,
            'supported': driver.supported,
            'options': driver.driver_options,
            }

    return info


def output_dict():
    """Output the results as a JSON dict."""

    driver_list = []
    drivers = util.get_volume_drivers()
    for driver in drivers:
        driver_list.append(collect_driver_info(driver))

    print(json.dumps(driver_list))


def main():
    tools_dir = os.path.dirname(os.path.abspath(__file__))
    cinder_root = os.path.dirname(tools_dir)
    cur_dir = os.getcwd()
    os.chdir(cinder_root)
    args = parser.parse_args()

    try:
        if args.format == 'str':
            output_str(cinder_root, args)
        elif args.format == 'dict':
            output_dict()

    finally:
        os.chdir(cur_dir)


if __name__ == '__main__':
    main()



# Copyright 2014 IBM Corp.
# Copyright 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder import utils


class VersionedMethod(utils.ComparableMixin):

    def __init__(self, name, start_version, end_version, experimental, func):
        """Versioning information for a single method.

        Minimum and maximums are inclusive.

        :param name: Name of the method
        :param start_version: Minimum acceptable version
        :param end_version: Maximum acceptable_version
        :param func: Method to call
        """
        self.name = name
        self.start_version = start_version
        self.end_version = end_version
        self.experimental = experimental
        self.func = func

    def __str__(self):
        args = {
            'name': self.name,
            'start': self.start_version,
            'end': self.end_version
        }
        return ("Version Method %(name)s: min: %(start)s, max: %(end)s" % args)

    def _cmpkey(self):
        """Return the value used by ComparableMixin for rich comparisons."""
        return self.start_version



#    (c) Copyright 2019 Brocade, a Broadcom Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""
Brocade south bound connector to communicate with switch using
HTTP or HTTPS protocol.
"""

import requests
import six
import time

from oslo_log import log as logging
from oslo_serialization import base64
from oslo_utils import encodeutils

from cinder.i18n import _
from cinder.zonemanager.drivers.brocade import exception as b_exception
import cinder.zonemanager.drivers.brocade.fc_zone_constants as zone_constant


LOG = logging.getLogger(__name__)


class BrcdHTTPFCZoneClient(object):

    def __init__(self, ipaddress, username,
                 password, port, vfid, protocol):
        """Initializing the client with the parameters passed.

        Creates authentication token and authenticate with switch
        to ensure the credentials are correct and change the VF context.

        :param ipaddress: IP Address of the device.
        :param username: User id to login.
        :param password: User password.
        :param port: Device Communication port
        :param vfid: Virtual Fabric ID.
        :param protocol: Communication Protocol.
        """
        self.switch_ip = ipaddress
        self.switch_user = username
        self.switch_pwd = password
        self.protocol = protocol
        self.vfid = vfid
        self.cfgs = {}
        self.zones = {}
        self.alias = {}
        self.qlps = {}
        self.ifas = {}
        self.active_cfg = ''
        self.parsed_raw_zoneinfo = ""
        self.random_no = ''
        self.auth_version = ''
        self.session = None

        # Create and assign the authentication header based on the credentials
        self.auth_header = self.create_auth_token()

        # Authenticate with the switch
        # If authenticated successfully, save the auth status and
        # create auth header for future communication with the device.
        self.is_auth, self.auth_header = self.authenticate()
        self.check_change_vf_context()

    def connect(self, requestType, requestURL, payload='', header=None):
        """Connect to the switch using HTTP/HTTPS protocol.

        :param requestType: Connection Request method
        :param requestURL: Connection URL
        :param payload: Data to send with POST request
        :param header: Request Headers

        :returns: HTTP response data
        :raises BrocadeZoningHttpException:
        """
        try:
            if header is None:
                header = {}
            header.update({"User-Agent": "OpenStack Zone Driver"})

            # Ensure only one connection is made throughout the life cycle
            protocol = zone_constant.HTTP
            if self.protocol == zone_constant.PROTOCOL_HTTPS:
                protocol = zone_constant.HTTPS
            if self.session is None:
                self.session = requests.Session()
                adapter = requests.adapters.HTTPAdapter(pool_connections=1,
                                                        pool_maxsize=1)
                self.session.mount(protocol + '://', adapter)
            url = '%s://%s%s' % (protocol, self.switch_ip, requestURL)
            response = None
            if requestType == zone_constant.GET_METHOD:
                response = self.session.get(url,
                                            headers=(header),
                                            verify=False)
            elif requestType == zone_constant.POST_METHOD:
                response = self.session.post(url,
                                             payload,
                                             headers=(header),
                                             verify=False)

            # Throw exception when response status is not OK
            if response.status_code != zone_constant.STATUS_OK:
                msg = _("Error while querying page %(url)s on the switch, "
                        "reason %(error)s.") % {'url': url,
                                                'error': response.reason}
                raise b_exception.BrocadeZoningHttpException(msg)
            else:
                return response.text
        except requests.exceptions.ConnectionError as e:
            msg = (_("Error while connecting the switch %(switch_id)s "
                     "with protocol %(protocol)s. Error: %(error)s.")
                   % {'switch_id': self.switch_ip,
                      'protocol': self.protocol,
                      'error': six.text_type(e)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        except b_exception.BrocadeZoningHttpException as ex:
            msg = (_("Unexpected status code from the switch %(switch_id)s "
                     "with protocol %(protocol)s for url %(page)s. "
                     "Error: %(error)s")
                   % {'switch_id': self.switch_ip,
                      'protocol': self.protocol,
                      'page': requestURL,
                      'error': six.text_type(ex)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def create_auth_token(self):
        """Create the authentication token.

        Creates the authentication token to use in the authentication header
        return authentication header (Base64(username:password:random no)).

        :returns: Authentication Header
        :raises BrocadeZoningHttpException:
        """
        try:
            # Send GET request to secinfo.html to get random number
            response = self.connect(zone_constant.GET_METHOD,
                                    zone_constant.SECINFO_PAGE)
            parsed_data = self.get_parsed_data(response,
                                               zone_constant.SECINFO_BEGIN,
                                               zone_constant.SECINFO_END)

            # Get the auth version for 8.1.0b+ switches
            self.auth_version = self.get_nvp_value(parsed_data,
                                                   zone_constant.AUTHVERSION)

            if self.auth_version == "1":
                # Extract the random no from secinfo.html response
                self.random_no = self.get_nvp_value(parsed_data,
                                                    zone_constant.RANDOM)
                # Form the authentication string
                auth_string = '%s:%s:%s' % (self.switch_user, self.switch_pwd,
                                            self.random_no)
            else:
                auth_string = '%s:%s' % (self.switch_user, self.switch_pwd)
            auth_token = base64.encode_as_text(auth_string).strip()
            auth_header = (zone_constant.AUTH_STRING +
                           auth_token)  # Build the proper header
        except Exception as e:
            msg = (_("Error while creating authentication token: %s")
                   % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return auth_header

    def authenticate(self):
        """Authenticate with the switch.

        Returns authentication status with modified authentication
        header (Base64(username:xxx:random no)).

        :returns: Authentication status
        :raises BrocadeZoningHttpException:
        """
        headers = {zone_constant.AUTH_HEADER: self.auth_header}
        try:
            # GET Request to authenticate.html to verify the credentials
            response = self.connect(zone_constant.GET_METHOD,
                                    zone_constant.AUTHEN_PAGE,
                                    header=headers)
            parsed_data = self.get_parsed_data(response,
                                               zone_constant.AUTHEN_BEGIN,
                                               zone_constant.AUTHEN_END)
            isauthenticated = self.get_nvp_value(
                parsed_data, zone_constant.AUTHENTICATED)
            if isauthenticated == "yes":
                if self.auth_version == "3":
                    auth_id = self.get_nvp_value(parsed_data,
                                                 zone_constant.IDENTIFIER)
                    auth_string = '%s:xxx:%s' % (self.switch_user, auth_id)
                else:
                    # Replace password in the authentication string with xxx
                    auth_string = '%s:xxx:%s' % (self.switch_user,
                                                 self.random_no)
                auth_token = base64.encode_as_text(auth_string).strip()
                auth_header = zone_constant.AUTH_STRING + auth_token
                return True, auth_header
            else:
                auth_error_code = self.get_nvp_value(parsed_data, "errCode")
                msg = (_("Authentication failed, verify the switch "
                         "credentials, error code %s.") % auth_error_code)
                LOG.error(msg)
                raise b_exception.BrocadeZoningHttpException(reason=msg)
        except Exception as e:
            msg = (_("Error while authenticating with switch: %s.")
                   % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def get_session_info(self):
        """Get the session information from the switch

        :returns: Connection status information.
        """
        try:
            headers = {zone_constant.AUTH_HEADER: self.auth_header}
            # GET request to session.html
            response = self.connect(zone_constant.GET_METHOD,
                                    zone_constant.SESSION_PAGE_ACTION,
                                    header=headers)
        except Exception as e:
            msg = (_("Error while getting session information %s.")
                   % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return response

    def get_parsed_data(self, data, delim1, delim2):
        """Return the sub string between the delimiters.

        :param data: String to manipulate
        :param delim1: Delimiter 1
        :param delim2: Delimiter 2
        :returns: substring between the delimiters
        """
        try:
            start = data.index(delim1)
            start = start + len(delim1)
            end = data.index(delim2)
            return data[start:end]
        except ValueError as e:
            msg = (_("Error while parsing the data: %s.") % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def get_nvp_value(self, data, keyname):
        """Get the value for the key passed.

        :param data: NVP to manipulate
        :param keyname: Key name
        :returns: value for the NVP
        """
        try:
            start = data.index(keyname)
            start = start + len(keyname)
            temp = data[start:]
            end = temp.index("\n")
            return (temp[:end].lstrip('= '))
        except ValueError as e:
            msg = (_("Error while getting nvp value: %s.") % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def get_managable_vf_list(self, session_info):
        """List of VFIDs that can be managed.

        :param session_info: Session information from the switch
        :returns: manageable VF list
        :raises BrocadeZoningHttpException:
        """
        try:
            # Check the value of manageableLFList NVP,
            # throw exception as not supported if the nvp not available
            vf_list = self.get_nvp_value(session_info,
                                         zone_constant.MANAGEABLE_VF)
            if vf_list:
                vf_list = vf_list.split(",")  # convert the string to list
        except b_exception.BrocadeZoningHttpException as e:
            msg = (_("Error while checking whether "
                     "VF is available for management %s.") % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return vf_list[:-1]

    def change_vf_context(self, vfid, session_data):
        """Change the VF context in the session.

        :param vfid: VFID to which context should be changed.
        :param session_data: Session information from the switch
        :raises BrocadeZoningHttpException:
        """
        try:
            managable_vf_list = self.get_managable_vf_list(session_data)
            LOG.debug("Manageable VF IDs are %(vflist)s.",
                      {'vflist': managable_vf_list})
            # proceed changing the VF context
            # if VF id can be managed if not throw exception
            if vfid in managable_vf_list:
                headers = {zone_constant.AUTH_HEADER: self.auth_header}
                data = zone_constant.CHANGE_VF.format(vfid=vfid)
                response = self.connect(zone_constant.POST_METHOD,
                                        zone_constant.SESSION_PAGE,
                                        data,
                                        headers)
                parsed_info = self.get_parsed_data(response,
                                                   zone_constant.SESSION_BEGIN,
                                                   zone_constant.SESSION_END)
                session_LF_Id = self.get_nvp_value(parsed_info,
                                                   zone_constant.SESSION_LF_ID)
                if session_LF_Id == vfid:
                    LOG.info("VF context is changed in the session.")
                else:
                    msg = _("Cannot change VF context in the session.")
                    LOG.error(msg)
                    raise b_exception.BrocadeZoningHttpException(reason=msg)

            else:
                msg = (_("Cannot change VF context, "
                         "specified VF is not available "
                         "in the manageable VF list %(vf_list)s.")
                       % {'vf_list': managable_vf_list})
                LOG.error(msg)
                raise b_exception.BrocadeZoningHttpException(reason=msg)
        except b_exception.BrocadeZoningHttpException as e:
            msg = (_("Error while changing VF context %s.") % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def get_zone_info(self):
        """Parse all the zone information and store it in the dictionary."""

        try:
            self.cfgs = {}
            self.zones = {}
            self.active_cfg = ''
            self.alias = {}
            self.qlps = {}
            self.ifas = {}
            headers = {zone_constant.AUTH_HEADER: self.auth_header}
            # GET request to gzoneinfo.htm
            response = self.connect(zone_constant.GET_METHOD,
                                    zone_constant.ZONE_PAGE,
                                    header=headers)
            # get the zone string from the response
            self.parsed_raw_zoneinfo = self.get_parsed_data(
                response,
                zone_constant.ZONEINFO_BEGIN,
                zone_constant.ZONEINFO_END).strip("\n")
            LOG.debug("Original zone string from the switch: %(zoneinfo)s",
                      {'zoneinfo': self.parsed_raw_zoneinfo})
            # convert the zone string to list
            zoneinfo = self.parsed_raw_zoneinfo.split()
            i = 0
            while i < len(zoneinfo):
                info = zoneinfo[i]
                # check for the cfg delimiter
                if zone_constant.CFG_DELIM in info:
                    # extract the cfg name
                    cfg_name = info.lstrip(zone_constant.CFG_DELIM)
                    # update the dict as
                    # self.cfgs={cfg_name:zone_name1;zone_name2}
                    self.cfgs.update({cfg_name: zoneinfo[i + 1]})
                    i = i + 2
                # check for the zone delimiter
                elif zone_constant.ZONE_DELIM in info:
                    # extract the zone name
                    zone_name = info.lstrip(zone_constant.ZONE_DELIM)
                    # update the dict as
                    # self.zones={zone_name:members1;members2}
                    self.zones.update({zone_name: zoneinfo[i + 1]})
                    i = i + 2
                elif zone_constant.ALIAS_DELIM in info:
                    alias_name = info.lstrip(zone_constant.ALIAS_DELIM)
                    # update the dict as
                    # self.alias={alias_name:members1;members2}
                    self.alias.update({alias_name: zoneinfo[i + 1]})
                    i = i + 2
                # check for quickloop zones
                elif zone_constant.QLP_DELIM in info:
                    qlp_name = info.lstrip(zone_constant.QLP_DELIM)
                    # update the map as self.qlps={qlp_name:members1;members2}
                    self.qlps.update({qlp_name: zoneinfo[i + 1]})
                    i = i + 2
                # check for fabric assist zones
                elif zone_constant.IFA_DELIM in info:
                    ifa_name = info.lstrip(zone_constant.IFA_DELIM)
                    # update the map as self.ifas={ifa_name:members1;members2}
                    self.ifas.update({ifa_name: zoneinfo[i + 1]})
                    i = i + 2
                elif zone_constant.ACTIVE_CFG_DELIM in info:
                    # update the string self.active_cfg=cfg_name
                    self.active_cfg = info.lstrip(
                        zone_constant.ACTIVE_CFG_DELIM)
                    if self.active_cfg == zone_constant.DEFAULT_CFG:
                        self.active_cfg = ""
                    i = i + 2
                else:
                    i = i + 1
        except Exception as e:
            msg = (_("Error while changing VF context %s.") % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def is_supported_firmware(self):
        """Check firmware version is v6.4 or higher.

        This API checks if the firmware version per the plug-in support level.
        This only checks major and minor version.

        :returns: True if firmware is supported else False.
        :raises BrocadeZoningHttpException:
        """

        isfwsupported = False

        try:
            headers = {zone_constant.AUTH_HEADER: self.auth_header}
            # GET request to switch.html
            response = self.connect(zone_constant.GET_METHOD,
                                    zone_constant.SWITCH_PAGE,
                                    header=headers)
            parsed_data = self.get_parsed_data(response,
                                               zone_constant.SWITCHINFO_BEGIN,
                                               zone_constant.SWITCHINFO_END)

            # get the firmware version nvp value
            fwVersion = self.get_nvp_value(
                parsed_data,
                zone_constant.FIRMWARE_VERSION).lstrip('v')

            ver = fwVersion.split(".")
            LOG.debug("Firmware version: %(version)s.", {'version': ver})
            if int(ver[0] + ver[1]) > 63:
                isfwsupported = True

        except Exception as e:
            msg = (_("Error while checking the firmware version %s.")
                   % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return isfwsupported

    def get_active_zone_set(self):
        """Return the active zone configuration.

        Return active zoneset from fabric. When none of the configurations
        are active then it will return empty map.

        :returns: Map -- active zone set map in the following format

        .. code-block:: python

            {
                'zones':
                    {'openstack50060b0000c26604201900051ee8e329':
                        ['50060b0000c26604', '201900051ee8e329']
                    },
                'active_zone_config': 'OpenStack_Cfg'
            }

        :raises BrocadeZoningHttpException:
        """
        active_zone_set = {}
        zones_map = {}
        try:
            self.get_zone_info()  # get the zone information of the switch
            if self.active_cfg != '':
                # get the zones list of the active_Cfg
                zones_list = self.cfgs[self.active_cfg].split(";")
                for n in zones_list:
                    # build the zones map
                    zones_map.update(
                        {n: self.zones[n].split(";")})
            # Format map in the correct format
            active_zone_set = {
                "active_zone_config": self.active_cfg, "zones": zones_map}
            return active_zone_set
        except Exception as e:
            msg = (_("Failed getting active zone set from fabric %s.")
                   % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def add_zones(self, add_zones_info, activate, active_zone_set=None):
        """Add zone configuration.

        This method will add the zone configuration passed by user.

        :param add_zones_info: Zone names mapped to members. Zone members
                               are colon separated but case-insensitive

        .. code-block:: python

            {   zonename1:[zonememeber1,zonemember2,...],
                zonename2:[zonemember1, zonemember2,...]...}

            e.g:

            {
                'openstack50060b0000c26604201900051ee8e329':
                    ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']
            }

        :param activate: True will activate the zone config.
        :param active_zone_set: Active zone set dict retrieved from
                                get_active_zone_set method
        :raises BrocadeZoningHttpException:
        """
        LOG.debug("Add zones - zones passed: %(zones)s.",
                  {'zones': add_zones_info})
        cfg_name = zone_constant.CFG_NAME
        cfgs = self.cfgs
        zones = self.zones
        alias = self.alias
        qlps = self.qlps
        ifas = self.ifas
        active_cfg = self.active_cfg
        # update the active_cfg, zones and cfgs map with new information
        zones, cfgs, active_cfg = self.add_zones_cfgs(cfgs,
                                                      zones,
                                                      add_zones_info,
                                                      active_cfg,
                                                      cfg_name)
        # Build the zonestring with updated maps
        data = self.form_zone_string(cfgs,
                                     active_cfg,
                                     zones,
                                     alias,
                                     qlps,
                                     ifas,
                                     activate)
        LOG.debug("Add zones: final zone string after applying "
                  "to the switch: %(zonestring)s", {'zonestring': data})
        # Post the zone data to the switch
        error_code, error_msg = self.post_zone_data(data)
        if error_code != "0":
            msg = (_("Applying the zones and cfgs to the switch failed "
                     "(error code=%(err_code)s error msg=%(err_msg)s.")
                   % {'err_code': error_code, 'err_msg': error_msg})

            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def update_zones(self, zone_info, activate, operation,
                     active_zone_set=None):
        """Update zone configuration.

        This method will update the zone configuration passed by user.

        :param zone_info: Zone names mapped to members. Zone members
                          are colon separated but case-insensitive

        .. code-block:: python

            {   zonename1:[zonememeber1,zonemember2,...],
                zonename2:[zonemember1, zonemember2,...]...}

            e.g:

            {
                'openstack50060b0000c26604201900051ee8e329':
                    ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']
            }

        :param activate: True will activate the zone config.
        :param operation: ZONE_ADD or ZONE_REMOVE
        :param active_zone_set: Active zone set dict retrieved from
                                get_active_zone_set method
        :raises BrocadeZoningHttpException:
        """
        LOG.debug("Update zones - zones passed: %(zones)s.",
                  {'zones': zone_info})
        cfgs = self.cfgs
        zones = self.zones
        alias = self.alias
        qlps = self.qlps
        ifas = self.ifas
        active_cfg = self.active_cfg
        # update the zones with new information
        zones = self._update_zones(zones, zone_info, operation)
        # Build the zonestring with updated maps
        data = self.form_zone_string(cfgs,
                                     active_cfg,
                                     zones,
                                     alias,
                                     qlps,
                                     ifas,
                                     activate)
        LOG.debug("Update zones: final zone string after applying "
                  "to the switch: %(zonestring)s", {'zonestring': data})
        # Post the zone data to the switch
        error_code, error_msg = self.post_zone_data(data)
        if error_code != "0":
            msg = (_("Applying the zones and cfgs to the switch failed "
                     "(error code=%(err_code)s error msg=%(err_msg)s.")
                   % {'err_code': error_code, 'err_msg': error_msg})

            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def form_zone_string(self, cfgs, active_cfg,
                         zones, alias, qlps, ifas, activate):
        """Build the zone string in the required format.

        :param cfgs:  cfgs map
        :param active_cfg: Active cfg string
        :param zones: zones map
        :param alias: alias map
        :param qlps: qlps map
        :param ifas: ifas map
        :param activate: True will activate config.
        :returns: zonestring in the required format
        :raises BrocadeZoningHttpException:
        """
        try:
            zoneString = zone_constant.ZONE_STRING_PREFIX

            # based on the activate save only will be changed
            saveonly = "false" if activate is True else "true"

            # Form the zone string based on the dictionary of each items
            for cfg in sorted(cfgs.keys()):
                zoneString += (zone_constant.CFG_DELIM +
                               cfg + " " + cfgs.get(cfg) + " ")
            for zone in sorted(zones.keys()):
                zoneString += (zone_constant.ZONE_DELIM +
                               zone + " " + zones.get(zone) + " ")
            for al in sorted(alias.keys()):
                zoneString += (zone_constant.ALIAS_DELIM +
                               al + " " + alias.get(al) + " ")
            for qlp in sorted(qlps.keys()):
                zoneString += (zone_constant.QLP_DELIM +
                               qlp + " " + qlps.get(qlp) + " ")
            for ifa in sorted(ifas.keys()):
                zoneString += (zone_constant.IFA_DELIM +
                               ifa + " " + ifas.get(ifa) + " ")
            # append the active_cfg string only if it is not null and activate
            # is true
            if active_cfg != "" and activate:
                zoneString += (zone_constant.ACTIVE_CFG_DELIM +
                               active_cfg + " null ")
            # Build the final zone string
            zoneString += zone_constant.ZONE_END_DELIM + saveonly
        except Exception as e:
            msg = (_("Exception while forming the zone string: %s.")
                   % six.text_type(e))
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        # Reconstruct the zoneString to type base string for OpenSSL
        return encodeutils.safe_encode(zoneString)

    def add_zones_cfgs(self, cfgs, zones, add_zones_info,
                       active_cfg, cfg_name):
        """Add the zones and cfgs map based on the new zones info.

        This method will return the updated zones,cfgs and active_cfg

        :param cfgs: Existing cfgs map
        :param active_cfg: Existing Active cfg string
        :param zones: Existing zones map
        :param add_zones_info: Zones map to add
        :param active_cfg: Existing active cfg
        :param cfg_name: New cfg name
        :returns: updated zones, zone configs map, and active_cfg
        """
        cfg_string = ""
        delimiter = ""
        zones_in_active_cfg = ""
        try:
            if active_cfg:
                zones_in_active_cfg = cfgs.get(active_cfg)
            for zone_name, members in add_zones_info.items():
                # if new zone is not active_cfg, build the cfg string with the
                # new zones
                if zone_name not in zones_in_active_cfg:
                    cfg_string += delimiter + zone_name
                    delimiter = ";"
                    # add a new zone with the members
                    zones.update({zone_name: ";".join(members)})
            # update cfg string
            if active_cfg:
                if cfg_string:
                    # update the existing active cfg map with cfgs string
                    cfgs.update(
                        {active_cfg: cfg_string + ";" + cfgs.get(active_cfg)})
            else:
                # create new cfg and update that cfgs map with the new cfg
                active_cfg = cfg_name
                cfgs.update({cfg_name: cfg_string})
        except Exception as e:
            msg = (_("Error while updating the new zones and cfgs "
                     "in the zone string. Error %(description)s.")
                   % {'description': six.text_type(e)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return zones, cfgs, active_cfg

    def _update_zones(self, zones, updated_zones_info, operation):
        """Update the zones based on the updated zones info.

        This method will return the updated zones

        :param zones: Existing zones map
        :param updated_zones_info: Zones map to update
        :param operation: ZONE_ADD or ZONE_REMOVE
        :returns: updated zones
        """
        try:
            for zone_name in updated_zones_info:
                members = updated_zones_info[zone_name]
                # update the zone string
                # if zone name already exists and dont have the new members
                # already
                current_members = zones.get(zone_name).split(";")
                if operation == zone_constant.ZONE_ADD:
                    new_members = set(members).difference(set(current_members))
                    if new_members:
                        # update the existing zone with new members
                        zones.update({zone_name: (";".join(new_members) +
                                     ";" + zones.get(zone_name))})
                else:
                    new_members = set(current_members).difference(set(members))
                    if new_members:
                        zones.pop(zone_name)
                        zones.update({zone_name: ";".join(new_members)})
        except Exception as e:
            msg = (_("Error while updating the zones "
                     "in the zone string. Error %(description)s.")
                   % {'description': six.text_type(e)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return zones

    def is_vf_enabled(self):
        """To check whether VF is enabled or not.

        :returns: boolean to indicate VF enabled and session information
        """
        session_info = self.get_session_info()
        parsed_data = self.get_parsed_data(session_info,
                                           zone_constant.SESSION_BEGIN,
                                           zone_constant.SESSION_END)
        try:
            is_vf_enabled = bool(self.get_nvp_value(
                parsed_data, zone_constant.VF_ENABLED))
        except b_exception.BrocadeZoningHttpException:
            is_vf_enabled = False
            parsed_data = None
        return is_vf_enabled, parsed_data

    def get_nameserver_info(self):
        """Get name server data from fabric.

        Return the connected node port wwn list(local
        and remote) for the given switch fabric.

        :returns: name server information.
        """
        nsinfo = []
        headers = {zone_constant.AUTH_HEADER: self.auth_header}
        response = self.connect(zone_constant.GET_METHOD,
                                zone_constant.NS_PAGE,
                                header=headers)  # GET request to nsinfo.html
        for line in response.splitlines():
            if line.startswith(zone_constant.NS_DELIM):
                nsinfo.append(line.split('=')[-1])
        return nsinfo

    def delete_zones_cfgs(
            self, cfgs, zones,
            delete_zones_info, active_cfg):
        """Delete the zones and cfgs map based on the new zones info.

        Return the updated zones, cfgs and active_cfg after deleting the
        required items.

        :param cfgs: Existing cfgs map
        :param active_cfg: Existing Active cfg string
        :param zones: Existing zones map
        :param delete_zones_info: Zones map to add
        :param active_cfg: Existing active cfg
        :returns: updated zones, zone config sets, and active zone config
        :raises BrocadeZoningHttpException:
        """
        try:
            delete_zones_info = delete_zones_info.split(";")
            for zone in delete_zones_info:
                # remove the zones from the zone map
                zones.pop(zone)
                # iterated all the cfgs, but need to check since in SSH only
                # active cfg is iterated
                for k, v in list(cfgs.items()):
                    v = v.split(";")
                    if zone in v:
                        # remove the zone from the cfg string
                        v.remove(zone)
                        # if all the zones are removed, remove the cfg from the
                        # cfg map
                        if not v:
                            cfgs.pop(k)
                        # update the original cfg with the updated string
                        else:
                            cfgs[k] = ";".join(v)

            # if all the zones are removed in the active_cfg, update it with
            # empty string
            if active_cfg not in cfgs:
                active_cfg = ""
        except KeyError as e:
            msg = (_("Error while removing the zones and cfgs "
                     "in the zone string: %(description)s.")
                   % {'description': six.text_type(e)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        return zones, cfgs, active_cfg

    def delete_zones(self, delete_zones_info, activate, active_zone_set=None):
        """Delete zones from fabric.

        Deletes zones in the active zone config.

        :param zone_names: zoneNames separated by semicolon
        :param activate: True/False
        :param active_zone_set: the active zone set dict retrieved
                                from get_active_zone_set method
        """
        cfgs = self.cfgs
        zones = self.zones
        alias = self.alias
        qlps = self.qlps
        ifas = self.ifas
        active_cfg = self.active_cfg
        # update the active_cfg, zones and cfgs map with required information
        # being removed
        zones, cfgs, active_cfg = self.delete_zones_cfgs(
            cfgs,
            zones,
            delete_zones_info,
            active_cfg)
        # Build the zonestring with updated maps
        data = self.form_zone_string(cfgs,
                                     active_cfg,
                                     zones,
                                     alias,
                                     qlps,
                                     ifas,
                                     activate)
        LOG.debug("Delete zones: final zone string after applying "
                  "to the switch: %(zonestring)s", {'zonestring': data})
        error_code, error_msg = self.post_zone_data(data)
        if error_code != "0":
            msg = (_("Applying the zones and cfgs to the switch failed "
                     "(error code=%(err_code)s error msg=%(err_msg)s.")
                   % {'err_code': error_code, 'err_msg': error_msg})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def post_zone_data(self, data):
        """Send POST request to the switch with the payload.

        :param data: payload to be sent to switch
        """

        status = "progress"
        parsed_data_txn = ""
        headers = {zone_constant.AUTH_HEADER: self.auth_header}

        LOG.debug("Requesting the switch with posting the zone string.")
        # POST request to gzoneinfo with zonestring as payload
        response = self.connect(zone_constant.POST_METHOD,
                                zone_constant.ZONE_PAGE,
                                data,
                                headers)
        parsed_data = self.get_parsed_data(response,
                                           zone_constant.ZONE_TX_BEGIN,
                                           zone_constant.ZONE_TX_END)
        transID = self.get_nvp_value(parsed_data,
                                     zone_constant.ZONE_TX_ID)
        transURL = zone_constant.ZONE_TRAN_STATUS.format(txnId=transID)
        timeout = 360
        sleep_time = 3
        time_elapsed = 0
        while(status != "done"):
            txn_response = self.connect(
                zone_constant.GET_METHOD, transURL, "", headers)
            parsed_data_txn = self.get_parsed_data(txn_response,
                                                   zone_constant.ZONE_TX_BEGIN,
                                                   zone_constant.ZONE_TX_END)
            status = self.get_nvp_value(parsed_data_txn,
                                        zone_constant.ZONE_TX_STATUS)
            time.sleep(sleep_time)
            time_elapsed += sleep_time
            if time_elapsed > timeout:
                break
        if status != "done":
            errorCode = -1
            errorMessage = ("Timed out, waiting for zone transaction on "
                            "the switch to complete")
        else:
            errorCode = self.get_nvp_value(parsed_data_txn,
                                           zone_constant.ZONE_ERROR_CODE)
            errorMessage = self.get_nvp_value(parsed_data_txn,
                                              zone_constant.ZONE_ERROR_MSG)
        return errorCode, errorMessage

    def check_change_vf_context(self):
        """Check whether VF related configurations is valid and proceed."""
        vf_enabled, session_data = self.is_vf_enabled()
        # VF enabled will be false if vf is disable or not supported
        LOG.debug("VF enabled on switch: %(vfenabled)s.",
                  {'vfenabled': vf_enabled})
        # Change the VF context in the session
        if vf_enabled:
            if self.vfid is None:
                msg = _("No VF ID is defined in the configuration file.")
                LOG.error(msg)
                raise b_exception.BrocadeZoningHttpException(reason=msg)
            elif self.vfid != 128:
                self.change_vf_context(self.vfid, session_data)
        else:
            if self.vfid is not None:
                msg = _("VF is not enabled.")
                LOG.error(msg)
                raise b_exception.BrocadeZoningHttpException(reason=msg)

    def _disconnect(self):
        """Disconnect from the switch using HTTP/HTTPS protocol.

        :raises BrocadeZoningHttpException:
        """
        try:
            headers = {zone_constant.AUTH_HEADER: self.auth_header}
            response = self.connect(zone_constant.GET_METHOD,
                                    zone_constant.LOGOUT_PAGE,
                                    header=headers)
            return response
        except requests.exceptions.ConnectionError as e:
            msg = (_("Error while connecting the switch %(switch_id)s "
                     "with protocol %(protocol)s. Error: %(error)s.")
                   % {'switch_id': self.switch_ip,
                      'protocol': self.protocol,
                      'error': six.text_type(e)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)
        except b_exception.BrocadeZoningHttpException as ex:
            msg = (_("Unexpected status code from the switch %(switch_id)s "
                     "with protocol %(protocol)s for url %(page)s. "
                     "Error: %(error)s")
                   % {'switch_id': self.switch_ip,
                      'protocol': self.protocol,
                      'page': zone_constant.LOGOUT_PAGE,
                      'error': six.text_type(ex)})
            LOG.error(msg)
            raise b_exception.BrocadeZoningHttpException(reason=msg)

    def cleanup(self):
        """Close session."""
        self._disconnect()
        self.session.close()



#    Copyright (c) 2013-2017 Dell Inc, or its subsidiaries.
#    Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import unittest

from eventlet import greenthread
import mock
from oslo_concurrency import processutils
import paramiko
import six

from cinder import context
from cinder import exception
from cinder import ssh_utils
from cinder import test
from cinder import utils
from cinder.volume import configuration as conf
from cinder.volume.drivers.dell_emc import ps


class PSSeriesISCSIDriverTestCase(test.TestCase):

    def setUp(self):
        super(PSSeriesISCSIDriverTestCase, self).setUp()
        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.san_is_local = False
        self.configuration.san_ip = "10.0.0.1"
        self.configuration.san_login = "foo"
        self.configuration.san_password = "bar"
        self.configuration.san_ssh_port = 16022
        self.configuration.san_thin_provision = True
        self.configuration.san_private_key = 'foo'
        self.configuration.ssh_min_pool_conn = 1
        self.configuration.ssh_max_pool_conn = 5
        self.configuration.ssh_conn_timeout = 30
        self.configuration.eqlx_pool = 'non-default'
        self.configuration.eqlx_group_name = 'group-0'
        self.configuration.eqlx_cli_max_retries = 5

        self.configuration.use_chap_auth = True
        self.configuration.chap_username = 'admin'
        self.configuration.chap_password = 'password'

        self.configuration.max_over_subscription_ratio = 1.0

        self.driver_stats_output = ['TotalCapacity: 111GB',
                                    'FreeSpace: 11GB',
                                    'VolumeReportedSpace: 80GB',
                                    'TotalVolumes: 100']
        self.cmd = 'this is dummy command'
        self._context = context.get_admin_context()
        self.driver = ps.PSSeriesISCSIDriver(
            configuration=self.configuration)
        self.volume_name = "fakevolume"
        self.volid = "fakeid"
        self.volume = {'name': self.volume_name,
                       'display_name': 'fake_display_name'}
        self.connector = {
            'ip': '10.0.0.2',
            'initiator': 'iqn.1993-08.org.debian:01:2227dab76162',
            'host': 'fakehost'}
        self.access_record_output = [
            "ID  Initiator       Ipaddress     AuthMethod UserName   Apply-To",
            "--- --------------- ------------- ---------- ---------- --------",
            "1   iqn.1993-08.org.debian:01:222 *.*.*.*       none        both",
            "       7dab76162"]
        self.fake_access_id = '1'
        self.fake_iqn = 'iqn.2003-10.com.equallogic:group01:25366:fakev'
        self.fake_iqn_return = ['iSCSI target name is %s.' % self.fake_iqn]
        self.fake_volume_output = ["Size: 5GB",
                                   "iSCSI Name: %s" % self.fake_iqn,
                                   "Description: "]
        self.fake_volume_info = {'size': 5.0,
                                 'iSCSI_Name': self.fake_iqn}
        self.driver._group_ip = '10.0.1.6'
        self.properties = {
            'target_discovered': True,
            'target_portal': '%s:3260' % self.driver._group_ip,
            'target_iqn': self.fake_iqn,
            'volume_id': 1,
            'discard': True}
        self._model_update = {
            'provider_location': "%s:3260,1 %s 0" % (self.driver._group_ip,
                                                     self.fake_iqn),
            'provider_auth': 'CHAP %s %s' % (
                self.configuration.chap_username,
                self.configuration.chap_password)
        }

    def _fake_get_iscsi_properties(self, volume):
        return self.properties

    def test_create_volume(self):
        volume = {'name': self.volume_name, 'size': 1}
        mock_attrs = {'args': ['volume', 'create', volume['name'],
                               "%sG" % (volume['size']), 'pool',
                               self.configuration.eqlx_pool,
                               'thin-provision']}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.return_value = self.fake_iqn_return
            model_update = self.driver.create_volume(volume)
            self.assertEqual(self._model_update, model_update)

    def test_delete_volume(self):
        volume = {'name': self.volume_name, 'size': 1}
        show_attrs = {'args': ['volume', 'select', volume['name'], 'show']}
        off_attrs = {'args': ['volume', 'select', volume['name'], 'offline']}
        delete_attrs = {'args': ['volume', 'delete', volume['name']]}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**show_attrs)
            mock_eql_execute.configure_mock(**off_attrs)
            mock_eql_execute.configure_mock(**delete_attrs)
            self.driver.delete_volume(volume)

    def test_delete_absent_volume(self):
        volume = {'name': self.volume_name, 'size': 1, 'id': self.volid}
        mock_attrs = {'args': ['volume', 'select', volume['name'], 'show']}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.side_effect = processutils.ProcessExecutionError(
                stdout='% Error ..... does not exist.\n')
            self.driver.delete_volume(volume)

    def test_ensure_export(self):
        volume = {'name': self.volume_name, 'size': 1}
        mock_attrs = {'args': ['volume', 'select', volume['name'], 'show']}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            self.driver.ensure_export({}, volume)

    def test_create_snapshot(self):
        snapshot = {'name': 'fakesnap', 'volume_name': 'fakevolume_name'}
        snap_name = 'fake_snap_name'
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.return_value = ['Snapshot name is %s' % snap_name]
            self.driver.create_snapshot(snapshot)

    def test_create_volume_from_snapshot(self):
        snapshot = {'name': 'fakesnap', 'volume_name': 'fakevolume_name',
                    'volume_size': '1'}
        volume = {'name': self.volume_name, 'size': '1'}
        mock_attrs = {'args': ['volume', 'select', snapshot['volume_name'],
                               'snapshot', 'select', snapshot['name'],
                               'clone', volume['name']]}

        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            with mock.patch.object(self.driver,
                                   'extend_volume') as mock_extend_volume:
                mock_eql_execute.configure_mock(**mock_attrs)
                mock_eql_execute.return_value = self.fake_iqn_return
                mock_extend_volume.return_value = self.fake_iqn_return
                model_update = self.driver.create_volume_from_snapshot(
                    volume, snapshot)
                self.assertEqual(self._model_update, model_update)
                self.assertFalse(self.driver.extend_volume.called)

    def test_create_volume_from_snapshot_extend(self):
        snapshot = {'name': 'fakesnap', 'volume_name': 'fakevolume_name',
                    'volume_size': '100'}
        volume = {'name': self.volume_name, 'size': '200'}
        mock_attrs = {'args': ['volume', 'select', snapshot['volume_name'],
                               'snapshot', 'select', snapshot['name'],
                               'clone', volume['name']]}

        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            with mock.patch.object(self.driver,
                                   'extend_volume') as mock_extend_volume:
                mock_eql_execute.configure_mock(**mock_attrs)
                mock_eql_execute.return_value = self.fake_iqn_return
                mock_extend_volume.return_value = self.fake_iqn_return
                model_update = self.driver.create_volume_from_snapshot(
                    volume, snapshot)
                self.assertEqual(self._model_update, model_update)
                self.assertTrue(self.driver.extend_volume.called)
                self.driver.extend_volume.assert_called_once_with(
                    volume, volume['size'])

    def test_create_cloned_volume(self):
        src_vref = {'name': 'fake_uuid', 'size': '1'}
        volume = {'name': self.volume_name, 'size': '1'}
        mock_attrs = {'args': ['volume', 'select', volume['name'],
                               'multihost-access', 'enable']}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            with mock.patch.object(self.driver,
                                   'extend_volume') as mock_extend_volume:
                mock_eql_execute.configure_mock(**mock_attrs)
                mock_eql_execute.return_value = self.fake_iqn_return
                mock_extend_volume.return_value = self.fake_iqn_return
                model_update = self.driver.create_cloned_volume(
                    volume, src_vref)
                self.assertEqual(self._model_update, model_update)
                self.assertFalse(self.driver.extend_volume.called)

    def test_create_cloned_volume_extend(self):
        src_vref = {'name': 'fake_uuid', 'size': '100'}
        volume = {'name': self.volume_name, 'size': '200'}
        mock_attrs = {'args': ['volume', 'select', volume['name'],
                               'multihost-access', 'enable']}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            with mock.patch.object(self.driver,
                                   'extend_volume') as mock_extend_volume:
                mock_eql_execute.configure_mock(**mock_attrs)
                mock_eql_execute.return_value = self.fake_iqn_return
                mock_extend_volume.return_value = self.fake_iqn_return
                cloned_vol = self.driver.create_cloned_volume(volume, src_vref)
                self.assertEqual(self._model_update, cloned_vol)
                self.assertTrue(self.driver.extend_volume.called)

    def test_delete_snapshot(self):
        snapshot = {'name': 'fakesnap', 'volume_name': 'fakevolume_name'}
        mock_attrs = {'args': ['volume', 'select', snapshot['volume_name'],
                               'snapshot', 'delete', snapshot['name']]}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            self.driver.delete_snapshot(snapshot)

    def test_delete_absent_snapshot(self):
        snapshot = {'name': 'fakesnap', 'volume_name': 'fakevolume_name'}
        mock_attrs = {'args': ['volume', 'select', snapshot['volume_name'],
                               'snapshot', 'delete', snapshot['name']]}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.side_effect = processutils.ProcessExecutionError(
                stdout='% Error ..... does not exist.\n')
            self.driver.delete_snapshot(snapshot)

    def test_extend_volume(self):
        new_size = '200'
        volume = {'name': self.volume_name, 'size': 100}
        mock_attrs = {'args': ['volume', 'select', volume['name'],
                               'size', "%sG" % new_size]}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            self.driver.extend_volume(volume, new_size)

    def test_get_volume_info(self):
        attrs = ('volume', 'select', self.volume, 'show')
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.return_value = self.fake_volume_output
            data = self.driver._get_volume_info(self.volume)
            mock_eql_execute.assert_called_with(*attrs)
            self.assertEqual(self.fake_volume_info, data)

    def test_get_volume_info_negative(self):
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.side_effect = processutils.ProcessExecutionError(
                stdout='% Error ..... does not exist.\n')
            self.assertRaises(exception.ManageExistingInvalidReference,
                              self.driver._get_volume_info, self.volume_name)

    def test_manage_existing(self):
        ref = {'source-name': self.volume_name}
        attrs = ('volume', 'select', self.volume_name,
                 'multihost-access', 'enable')
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            with mock.patch.object(self.driver,
                                   '_get_volume_info') as mock_volume_info:
                mock_volume_info.return_value = self.fake_volume_info
                mock_eql_execute.return_value = self.fake_iqn_return
                model_update = self.driver.manage_existing(self.volume, ref)
                mock_eql_execute.assert_called_with(*attrs)
                self.assertEqual(self._model_update, model_update)

    def test_manage_existing_invalid_ref(self):
        ref = {}
        self.assertRaises(exception.InvalidInput,
                          self.driver.manage_existing, self.volume, ref)

    def test_manage_existing_get_size(self):
        ref = {'source-name': self.volume_name}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.return_value = self.fake_volume_output
            size = self.driver.manage_existing_get_size(self.volume, ref)
            self.assertEqual(float('5.0'), size)

    def test_manage_existing_get_size_invalid_ref(self):
        """Error on manage with invalid reference."""
        ref = {}
        self.assertRaises(exception.InvalidInput,
                          self.driver.manage_existing_get_size,
                          self.volume, ref)

    def test_unmanage(self):
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.return_value = None
            self.driver.unmanage(self.volume)

    def test_initialize_connection(self):
        volume = {'name': self.volume_name}
        mock_attrs = {'args': ['volume', 'select', volume['name'], 'access',
                               'create', 'initiator',
                               self.connector['initiator'],
                               'authmethod', 'chap',
                               'username',
                               self.configuration.chap_username]}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            with mock.patch.object(self.driver,
                                   '_get_iscsi_properties') as mock_iscsi:
                mock_eql_execute.configure_mock(**mock_attrs)
                mock_iscsi.return_value = self.properties
                iscsi_properties = self.driver.initialize_connection(
                    volume, self.connector)
                self.assertEqual(self._fake_get_iscsi_properties(volume),
                                 iscsi_properties['data'])
                self.assertTrue(iscsi_properties['data']['discard'])

    def test_terminate_connection(self):
        def my_side_effect(*args, **kwargs):
            if args[4] == 'show':
                return self.access_record_output
            else:
                return ''
        volume = {'name': self.volume_name}
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.side_effect = my_side_effect
            self.driver.terminate_connection(volume, self.connector)

    def test_get_access_record(self):
        attrs = ('volume', 'select', self.volume['name'], 'access', 'show')
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.return_value = self.access_record_output
            data = self.driver._get_access_record(self.volume, self.connector)
            mock_eql_execute.assert_called_with(*attrs)
            self.assertEqual(self.fake_access_id, data)

    def test_get_access_record_negative(self):
        attrs = ('volume', 'select', self.volume['name'], 'access', 'show')
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.return_value = []
            data = self.driver._get_access_record(self.volume, self.connector)
            mock_eql_execute.assert_called_with(*attrs)
            self.assertIsNone(data)

    def test_do_setup(self):
        fake_group_ip = '10.1.2.3'

        def my_side_effect(*args, **kwargs):
            if args[0] == 'grpparams':
                return ['Group-Ipaddress: %s' % fake_group_ip]
            else:
                return ''

        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.side_effect = my_side_effect
            self.driver.do_setup(self._context)
            self.assertEqual(fake_group_ip, self.driver._group_ip)

    def test_update_volume_stats_thin(self):
        mock_attrs = {'args': ['pool', 'select',
                               self.configuration.eqlx_pool, 'show']}
        self.configuration.san_thin_provision = True
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.return_value = self.driver_stats_output
            self.driver._update_volume_stats()
            self.assert_volume_stats(self.driver._stats)

    def test_update_volume_stats_thick(self):
        mock_attrs = {'args': ['pool', 'select',
                               self.configuration.eqlx_pool, 'show']}
        self.configuration.san_thin_provision = False
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.return_value = self.driver_stats_output
            self.driver._update_volume_stats()
            self.assert_volume_stats(self.driver._stats)

    def test_get_volume_stats_thin(self):
        mock_attrs = {'args': ['pool', 'select',
                               self.configuration.eqlx_pool, 'show']}
        self.configuration.san_thin_provision = True
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.return_value = self.driver_stats_output
            stats = self.driver.get_volume_stats(refresh=True)
            self.assert_volume_stats(stats)

    def test_get_volume_stats_thick(self):
        mock_attrs = {'args': ['pool', 'select',
                               self.configuration.eqlx_pool, 'show']}
        self.configuration.san_thin_provision = False
        with mock.patch.object(self.driver,
                               '_eql_execute') as mock_eql_execute:
            mock_eql_execute.configure_mock(**mock_attrs)
            mock_eql_execute.return_value = self.driver_stats_output
            stats = self.driver.get_volume_stats(refresh=True)
            self.assert_volume_stats(stats)

    def assert_volume_stats(self, stats):
        thin_enabled = self.configuration.san_thin_provision
        self.assertEqual(float('111.0'), stats['total_capacity_gb'])
        self.assertEqual(float('11.0'), stats['free_capacity_gb'])
        self.assertEqual(100, stats['total_volumes'])

        if thin_enabled:
            self.assertEqual(80.0, stats['provisioned_capacity_gb'])
        else:
            space = stats['total_capacity_gb'] - stats['free_capacity_gb']
            self.assertEqual(space, stats['provisioned_capacity_gb'])

        self.assertEqual(thin_enabled, stats['thin_provisioning_support'])
        self.assertEqual(not thin_enabled,
                         stats['thick_provisioning_support'])
        self.assertEqual('Dell EMC', stats['vendor_name'])
        self.assertFalse(stats['multiattach'])

    def test_get_space_in_gb(self):
        self.assertEqual(123.0, self.driver._get_space_in_gb('123.0GB'))
        self.assertEqual(124.0, self.driver._get_space_in_gb('123.5GB'))
        self.assertEqual(123.0 * 1024, self.driver._get_space_in_gb('123.0TB'))
        self.assertEqual(1.0, self.driver._get_space_in_gb('1024.0MB'))
        self.assertEqual(2.0, self.driver._get_space_in_gb('1536.0MB'))

    def test_get_output(self):

        def _fake_recv(ignore_arg):
            return '%s> ' % self.configuration.eqlx_group_name

        chan = mock.Mock(paramiko.Channel)
        mock_recv = self.mock_object(chan, 'recv')
        mock_recv.return_value = '%s> ' % self.configuration.eqlx_group_name
        self.assertEqual([_fake_recv(None)], self.driver._get_output(chan))

    def test_get_prefixed_value(self):
        lines = ['Line1 passed', 'Line1 failed']
        prefix = ['Line1', 'Line2']
        expected_output = [' passed', None]
        self.assertEqual(expected_output[0],
                         self.driver._get_prefixed_value(lines, prefix[0]))
        self.assertEqual(expected_output[1],
                         self.driver._get_prefixed_value(lines, prefix[1]))

    def test_ssh_execute(self):
        ssh = mock.Mock(paramiko.SSHClient)
        chan = mock.Mock(paramiko.Channel)
        transport = mock.Mock(paramiko.Transport)
        mock_get_output = self.mock_object(self.driver, '_get_output')
        self.mock_object(chan, 'invoke_shell')
        expected_output = ['NoError: test run']
        mock_get_output.return_value = expected_output
        ssh.get_transport.return_value = transport
        transport.open_session.return_value = chan
        chan.invoke_shell()
        chan.send('stty columns 255' + '\r')
        chan.send(self.cmd + '\r')
        chan.close()
        self.assertEqual(expected_output,
                         self.driver._ssh_execute(ssh, self.cmd))

    def test_ssh_execute_error(self):
        self.mock_object(self.driver, '_ssh_execute',
                         side_effect=processutils.ProcessExecutionError)
        ssh = mock.Mock(paramiko.SSHClient)
        chan = mock.Mock(paramiko.Channel)
        transport = mock.Mock(paramiko.Transport)
        mock_get_output = self.mock_object(self.driver, '_get_output')
        self.mock_object(ssh, 'get_transport')
        self.mock_object(chan, 'invoke_shell')
        expected_output = ['Error: test run', '% Error']
        mock_get_output.return_value = expected_output
        ssh.get_transport().return_value = transport
        transport.open_session.return_value = chan
        chan.invoke_shell()
        chan.send('stty columns 255' + '\r')
        chan.send(self.cmd + '\r')
        chan.close()
        self.assertRaises(processutils.ProcessExecutionError,
                          self.driver._ssh_execute, ssh, self.cmd)

    @mock.patch.object(greenthread, 'sleep')
    def test_ensure_retries(self, _gt_sleep):
        num_attempts = 3
        self.driver.configuration.eqlx_cli_max_retries = num_attempts
        self.mock_object(self.driver, '_ssh_execute',
                         side_effect=exception.VolumeBackendAPIException(
                             "some error"))
        # mocks for calls in _run_ssh
        self.mock_object(utils, 'check_ssh_injection')
        self.mock_object(ssh_utils, 'SSHPool')

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)
        self.driver.sshpool = mock.Mock(return_value=sshpool)
        ssh = mock.Mock(paramiko.SSHClient)
        self.driver.sshpool.item().__enter__ = mock.Mock(return_value=ssh)
        self.driver.sshpool.item().__exit__ = mock.Mock(return_value=False)
        # now call the execute
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver._eql_execute, "fake command")
        self.assertEqual(num_attempts + 1,
                         self.driver._ssh_execute.call_count)

    @mock.patch.object(greenthread, 'sleep')
    def test_ensure_connection_retries(self, _gt_sleep):
        num_attempts = 3
        self.driver.configuration.eqlx_cli_max_retries = num_attempts
        self.mock_object(self.driver, '_ssh_execute',
                         side_effect=processutils.ProcessExecutionError(
                             stdout='% Error ... some error.\n'))
        # mocks for calls in _run_ssh
        self.mock_object(utils, 'check_ssh_injection')
        self.mock_object(ssh_utils, 'SSHPool')

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)
        self.driver.sshpool = mock.Mock(return_value=sshpool)
        ssh = mock.Mock(paramiko.SSHClient)
        self.driver.sshpool.item().__enter__ = mock.Mock(return_value=ssh)
        self.driver.sshpool.item().__exit__ = mock.Mock(return_value=False)
        # now call the execute
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver._eql_execute, "fake command")
        self.assertEqual(num_attempts + 1,
                         self.driver._ssh_execute.call_count)

    @unittest.skip("Skip until bug #1578986 is fixed")
    @mock.patch.object(greenthread, 'sleep')
    def test_ensure_retries_on_channel_timeout(self, _gt_sleep):
        num_attempts = 3
        self.driver.configuration.eqlx_cli_max_retries = num_attempts

        # mocks for calls and objects in _run_ssh
        self.mock_object(utils, 'check_ssh_injection')
        self.mock_object(ssh_utils, 'SSHPool')

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)
        self.driver.sshpool = mock.Mock(return_value=sshpool)
        ssh = mock.Mock(paramiko.SSHClient)
        self.driver.sshpool.item().__enter__ = mock.Mock(return_value=ssh)
        self.driver.sshpool.item().__exit__ = mock.Mock(return_value=False)
        # mocks for _ssh_execute and _get_output
        self.mock_object(self.driver, '_get_output',
                         side_effect=exception.VolumeBackendAPIException(
                             "some error"))
        # now call the execute
        with mock.patch('sys.stderr', new=six.StringIO()):
            self.assertRaises(exception.VolumeBackendAPIException,
                              self.driver._eql_execute, "fake command")

        self.assertEqual(num_attempts + 1, self.driver._get_output.call_count)

    @unittest.skip("Skip until bug #1578986 is fixed")
    def test_with_timeout(self):
        @ps.with_timeout
        def no_timeout(cmd, *args, **kwargs):
            return 'no timeout'

        @ps.with_timeout
        def w_timeout(cmd, *args, **kwargs):
            time.sleep(1)

        self.assertEqual('no timeout', no_timeout('fake cmd'))
        self.assertRaises(exception.VolumeBackendAPIException,
                          w_timeout, 'fake cmd', timeout=0.1)

    def test_local_path(self):
        self.assertRaises(NotImplementedError, self.driver.local_path, '')



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_utils import timeutils

import cinder.api.contrib.availability_zones
import cinder.context
import cinder.test
import cinder.volume.api


created_time = datetime.datetime(2012, 11, 14, 1, 20, 41, 95099)
current_time = timeutils.utcnow()


def list_availability_zones(self):
    return (
        {'name': 'ping', 'available': True},
        {'name': 'pong', 'available': False},
    )


class FakeRequest(object):
    environ = {'cinder.context': cinder.context.get_admin_context()}
    GET = {}


class ControllerTestCase(cinder.test.TestCase):

    def setUp(self):
        super(ControllerTestCase, self).setUp()
        self.controller = cinder.api.contrib.availability_zones.Controller()
        self.req = FakeRequest()
        self.mock_object(cinder.volume.api.API,
                         'list_availability_zones',
                         list_availability_zones)

    def test_list_hosts(self):
        """Verify that the volume hosts are returned."""
        actual = self.controller.index(self.req)
        expected = {
            'availabilityZoneInfo': [
                {'zoneName': 'ping', 'zoneState': {'available': True}},
                {'zoneName': 'pong', 'zoneState': {'available': False}},
            ],
        }
        self.assertEqual(expected, actual)



# Copyright 2012 Pedro Navarro Perez
# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for Windows Server 2012 OpenStack Cinder volume driver
"""

import os

import ddt
import mock
from oslo_utils import fileutils
from oslo_utils import units

from cinder import context
from cinder import exception
from cinder.image import image_utils
from cinder import test
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit.windows import db_fakes
from cinder.volume import configuration as conf
from cinder.volume.drivers.windows import iscsi as windows_iscsi


@ddt.ddt
class TestWindowsISCSIDriver(test.TestCase):
    @mock.patch.object(windows_iscsi, 'utilsfactory')
    def setUp(self, mock_utilsfactory):
        super(TestWindowsISCSIDriver, self).setUp()
        self.configuration = conf.Configuration(None)
        self.configuration.append_config_values(windows_iscsi.windows_opts)
        self.flags(windows_iscsi_lun_path='fake_iscsi_lun_path')
        self.flags(image_conversion_dir='fake_image_conversion_dir')

        self._driver = windows_iscsi.WindowsISCSIDriver(
            configuration=self.configuration)

    @mock.patch.object(fileutils, 'ensure_tree')
    def test_do_setup(self, mock_ensure_tree):
        self._driver.do_setup(mock.sentinel.context)

        mock_ensure_tree.assert_has_calls(
            [mock.call('fake_iscsi_lun_path'),
             mock.call('fake_image_conversion_dir')])

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_get_portals')
    def test_check_for_setup_error(self, mock_get_portals):
        self._driver.check_for_setup_error()

        mock_get_portals.assert_called_once_with()

    @ddt.data(True, False)
    def test_get_portals(self, portals_available=True):
        iscsi_port = mock.sentinel.iscsi_port
        available_ips = ['fake_ip0', 'fake_ip1', 'fake_unrequested_ip']
        requested_ips = available_ips[:-1] + ['fake_inexistent_ips']

        available_portals = ([":".join([ip_addr, str(iscsi_port)])
                              for ip_addr in available_ips]
                             if portals_available else [])

        self._driver.configuration = mock.Mock()
        self._driver.configuration.target_port = iscsi_port
        self._driver.configuration.target_ip_address = requested_ips[0]
        self._driver.configuration.iscsi_secondary_ip_addresses = (
            requested_ips[1:])

        self._driver._tgt_utils.get_portal_locations.return_value = (
            available_portals)

        if portals_available:
            portals = self._driver._get_portals()
            self.assertEqual(set(available_portals[:-1]), set(portals))
        else:
            self.assertRaises(exception.VolumeDriverException,
                              self._driver._get_portals)

        self._driver._tgt_utils.get_portal_locations.assert_called_once_with(
            available_only=True,
            fail_if_none_found=True)

    @ddt.data(True, False)
    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_get_portals')
    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_get_target_name')
    def test_get_host_information(self, multipath, mock_get_target_name,
                                  mock_get_portals):
        tgt_utils = self._driver._tgt_utils

        fake_auth_meth = 'CHAP'
        fake_chap_username = 'fake_chap_username'
        fake_chap_password = 'fake_chap_password'
        fake_target_iqn = 'fake_target_iqn'
        fake_host_info = {'target_iqn': 'fake_target_iqn',
                          'fake_prop': 'fake_value'}
        fake_provider_auth = "%s %s %s" % (fake_auth_meth,
                                           fake_chap_username,
                                           fake_chap_password)
        fake_portals = [mock.sentinel.portal_location0,
                        mock.sentinel.portal_location1]

        volume = fake_volume.fake_volume_obj(mock.sentinel.context,
                                             provider_auth=fake_provider_auth)

        mock_get_target_name.return_value = mock.sentinel.target_name
        mock_get_portals.return_value = fake_portals
        tgt_utils.get_target_information.return_value = fake_host_info

        expected_host_info = dict(fake_host_info,
                                  auth_method=fake_auth_meth,
                                  auth_username=fake_chap_username,
                                  auth_password=fake_chap_password,
                                  target_discovered=False,
                                  target_portal=fake_portals[0],
                                  target_lun=0,
                                  volume_id=volume.id)
        if multipath:
            expected_host_info['target_portals'] = fake_portals
            expected_host_info['target_iqns'] = [fake_target_iqn] * 2
            expected_host_info['target_luns'] = [0] * 2

        host_info = self._driver._get_host_information(volume, multipath)

        self.assertEqual(expected_host_info, host_info)

        mock_get_target_name.assert_called_once_with(volume)
        mock_get_portals.assert_called_once_with()
        tgt_utils.get_target_information.assert_called_once_with(
            mock.sentinel.target_name)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver,
                       '_get_host_information')
    def test_initialize_connection(self, mock_get_host_info):
        tgt_utils = self._driver._tgt_utils

        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        fake_initiator = db_fakes.get_fake_connector_info()
        fake_initiator['multipath'] = mock.sentinel.multipath
        fake_host_info = {'fake_host_prop': 'fake_value'}

        mock_get_host_info.return_value = fake_host_info

        expected_conn_info = {'driver_volume_type': 'iscsi',
                              'data': fake_host_info}
        conn_info = self._driver.initialize_connection(volume,
                                                       fake_initiator)

        self.assertEqual(expected_conn_info, conn_info)
        mock_get_host_info.assert_called_once_with(
            volume, mock.sentinel.multipath)
        mock_associate = tgt_utils.associate_initiator_with_iscsi_target
        mock_associate.assert_called_once_with(
            fake_initiator['initiator'],
            volume.provider_location)

    def test_terminate_connection(self):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        fake_initiator = db_fakes.get_fake_connector_info()

        self._driver.terminate_connection(volume, fake_initiator)

        self._driver._tgt_utils.deassociate_initiator.assert_called_once_with(
            fake_initiator['initiator'], volume.provider_location)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, 'local_path')
    def test_create_volume(self, mock_local_path):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)

        self._driver.create_volume(volume)

        mock_local_path.assert_called_once_with(volume)
        self._driver._tgt_utils.create_wt_disk.assert_called_once_with(
            mock_local_path.return_value,
            volume.name,
            size_mb=volume.size * 1024)

    def test_local_path(self):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)

        fake_lun_path = 'fake_lun_path'
        self.flags(windows_iscsi_lun_path=fake_lun_path)

        disk_format = 'vhd'
        mock_get_fmt = self._driver._tgt_utils.get_supported_disk_format
        mock_get_fmt.return_value = disk_format

        disk_path = self._driver.local_path(volume)

        expected_fname = "%s.%s" % (volume.name, disk_format)
        expected_disk_path = os.path.join(fake_lun_path,
                                          expected_fname)
        self.assertEqual(expected_disk_path, disk_path)
        mock_get_fmt.assert_called_once_with()

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, 'local_path')
    @mock.patch.object(fileutils, 'delete_if_exists')
    def test_delete_volume(self, mock_delete_if_exists, mock_local_path):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)

        self._driver.delete_volume(volume)

        mock_local_path.assert_called_once_with(volume)
        self._driver._tgt_utils.remove_wt_disk.assert_called_once_with(
            volume.name)
        mock_delete_if_exists.assert_called_once_with(
            mock_local_path.return_value)

    def test_create_snapshot(self):
        volume = fake_volume.fake_volume_obj(context.get_admin_context())
        snapshot = fake_snapshot.fake_snapshot_obj(context.get_admin_context(),
                                                   volume_id=volume.id)
        snapshot.volume = volume

        self._driver.create_snapshot(snapshot)

        self._driver._tgt_utils.create_snapshot.assert_called_once_with(
            snapshot.volume_name, snapshot.name)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, 'local_path')
    def test_create_volume_from_snapshot(self, mock_local_path):
        volume = fake_volume.fake_volume_obj(context.get_admin_context())
        snapshot = fake_snapshot.fake_snapshot_obj(context.get_admin_context())
        snapshot.volume = volume

        self._driver.create_volume_from_snapshot(volume, snapshot)

        self._driver._tgt_utils.export_snapshot.assert_called_once_with(
            snapshot.name, mock_local_path.return_value)
        self._driver._tgt_utils.import_wt_disk.assert_called_once_with(
            mock_local_path.return_value, volume.name)

    def test_delete_snapshot(self):
        snapshot = fake_snapshot.fake_snapshot_obj(context.get_admin_context())

        self._driver.delete_snapshot(snapshot)

        self._driver._tgt_utils.delete_snapshot.assert_called_once_with(
            snapshot.name)

    def test_get_target_name(self):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        expected_target_name = "%s%s" % (
            self._driver.configuration.target_prefix,
            volume.name)

        target_name = self._driver._get_target_name(volume)
        self.assertEqual(expected_target_name, target_name)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_get_target_name')
    @mock.patch.object(windows_iscsi.volume_utils, 'generate_username')
    @mock.patch.object(windows_iscsi.volume_utils, 'generate_password')
    def test_create_export(self, mock_generate_password,
                           mock_generate_username,
                           mock_get_target_name):
        tgt_utils = self._driver._tgt_utils
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        self._driver.configuration.chap_username = None
        self._driver.configuration.chap_password = None
        self._driver.configuration.use_chap_auth = True
        fake_chap_username = 'fake_chap_username'
        fake_chap_password = 'fake_chap_password'

        mock_get_target_name.return_value = mock.sentinel.target_name
        mock_generate_username.return_value = fake_chap_username
        mock_generate_password.return_value = fake_chap_password
        tgt_utils.iscsi_target_exists.return_value = False

        vol_updates = self._driver.create_export(mock.sentinel.context,
                                                 volume,
                                                 mock.sentinel.connector)

        mock_get_target_name.assert_called_once_with(volume)
        tgt_utils.iscsi_target_exists.assert_called_once_with(
            mock.sentinel.target_name)
        tgt_utils.set_chap_credentials.assert_called_once_with(
            mock.sentinel.target_name,
            fake_chap_username,
            fake_chap_password)
        tgt_utils.add_disk_to_target.assert_called_once_with(
            volume.name, mock.sentinel.target_name)

        expected_provider_auth = ' '.join(('CHAP',
                                           fake_chap_username,
                                           fake_chap_password))
        expected_vol_updates = dict(
            provider_location=mock.sentinel.target_name,
            provider_auth=expected_provider_auth)
        self.assertEqual(expected_vol_updates, vol_updates)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_get_target_name')
    def test_remove_export(self, mock_get_target_name):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)

        self._driver.remove_export(mock.sentinel.context, volume)

        mock_get_target_name.assert_called_once_with(volume)
        self._driver._tgt_utils.delete_iscsi_target.assert_called_once_with(
            mock_get_target_name.return_value)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, 'local_path')
    @mock.patch.object(image_utils, 'temporary_file')
    @mock.patch.object(image_utils, 'fetch_to_vhd')
    @mock.patch('os.unlink')
    def test_copy_image_to_volume(self, mock_unlink, mock_fetch_to_vhd,
                                  mock_tmp_file, mock_local_path):
        tgt_utils = self._driver._tgt_utils
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)

        mock_tmp_file.return_value.__enter__.return_value = (
            mock.sentinel.tmp_vhd_path)
        mock_local_path.return_value = mock.sentinel.vol_vhd_path

        self._driver.copy_image_to_volume(mock.sentinel.context,
                                          volume,
                                          mock.sentinel.image_service,
                                          mock.sentinel.image_id)

        mock_local_path.assert_called_once_with(volume)
        mock_tmp_file.assert_called_once_with(suffix='.vhd')
        image_utils.fetch_to_vhd.assert_called_once_with(
            mock.sentinel.context, mock.sentinel.image_service,
            mock.sentinel.image_id, mock.sentinel.tmp_vhd_path,
            self._driver.configuration.volume_dd_blocksize)

        mock_unlink.assert_called_once_with(mock.sentinel.vol_vhd_path)
        self._driver._vhdutils.convert_vhd.assert_called_once_with(
            mock.sentinel.tmp_vhd_path,
            mock.sentinel.vol_vhd_path,
            tgt_utils.get_supported_vhd_type.return_value)
        self._driver._vhdutils.resize_vhd.assert_called_once_with(
            mock.sentinel.vol_vhd_path,
            volume.size * units.Gi,
            is_file_max_size=False)

        tgt_utils.change_wt_disk_status.assert_has_calls(
            [mock.call(volume.name, enabled=False),
             mock.call(volume.name, enabled=True)])

    @mock.patch.object(windows_iscsi.uuidutils, 'generate_uuid')
    def test_temporary_snapshot(self, mock_generate_uuid):
        tgt_utils = self._driver._tgt_utils
        mock_generate_uuid.return_value = mock.sentinel.snap_uuid
        expected_snap_name = '%s-tmp-snapshot-%s' % (
            mock.sentinel.volume_name, mock.sentinel.snap_uuid)

        with self._driver._temporary_snapshot(
                mock.sentinel.volume_name) as snap_name:
            self.assertEqual(expected_snap_name, snap_name)
            tgt_utils.create_snapshot.assert_called_once_with(
                mock.sentinel.volume_name, expected_snap_name)

        tgt_utils.delete_snapshot.assert_called_once_with(
            expected_snap_name)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_temporary_snapshot')
    @mock.patch.object(image_utils, 'upload_volume')
    @mock.patch.object(fileutils, 'delete_if_exists')
    def test_copy_volume_to_image(self, mock_delete_if_exists,
                                  mock_upload_volume,
                                  mock_tmp_snap):
        tgt_utils = self._driver._tgt_utils

        disk_format = 'vhd'
        fake_image_meta = db_fakes.get_fake_image_meta()
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        fake_img_conv_dir = 'fake_img_conv_dir'
        self.flags(image_conversion_dir=fake_img_conv_dir)

        tgt_utils.get_supported_disk_format.return_value = disk_format
        mock_tmp_snap.return_value.__enter__.return_value = (
            mock.sentinel.tmp_snap_name)

        expected_tmp_vhd_path = os.path.join(
            fake_img_conv_dir,
            fake_image_meta['id'] + '.' + disk_format)

        self._driver.copy_volume_to_image(
            mock.sentinel.context, volume,
            mock.sentinel.image_service,
            fake_image_meta)

        mock_tmp_snap.assert_called_once_with(volume.name)
        tgt_utils.export_snapshot.assert_called_once_with(
            mock.sentinel.tmp_snap_name,
            expected_tmp_vhd_path)
        mock_upload_volume.assert_called_once_with(
            mock.sentinel.context, mock.sentinel.image_service,
            fake_image_meta, expected_tmp_vhd_path, 'vhd')
        mock_delete_if_exists.assert_called_once_with(
            expected_tmp_vhd_path)

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_temporary_snapshot')
    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, 'local_path')
    def test_create_cloned_volume(self, mock_local_path,
                                  mock_tmp_snap):
        tgt_utils = self._driver._tgt_utils

        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        src_volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)

        mock_tmp_snap.return_value.__enter__.return_value = (
            mock.sentinel.tmp_snap_name)
        mock_local_path.return_value = mock.sentinel.vol_vhd_path

        self._driver.create_cloned_volume(volume, src_volume)

        mock_tmp_snap.assert_called_once_with(src_volume.name)
        tgt_utils.export_snapshot.assert_called_once_with(
            mock.sentinel.tmp_snap_name,
            mock.sentinel.vol_vhd_path)
        self._driver._vhdutils.resize_vhd.assert_called_once_with(
            mock.sentinel.vol_vhd_path, volume.size * units.Gi,
            is_file_max_size=False)
        tgt_utils.import_wt_disk.assert_called_once_with(
            mock.sentinel.vol_vhd_path, volume.name)

    @mock.patch('os.path.splitdrive')
    def test_get_capacity_info(self, mock_splitdrive):
        mock_splitdrive.return_value = (mock.sentinel.drive,
                                        mock.sentinel.path_tail)
        fake_size_gb = 2
        fake_free_space_gb = 1
        self._driver._hostutils.get_volume_info.return_value = (
            fake_size_gb * units.Gi,
            fake_free_space_gb * units.Gi)

        total_gb, free_gb = self._driver._get_capacity_info()

        self.assertEqual(fake_size_gb, total_gb)
        self.assertEqual(fake_free_space_gb, free_gb)

        self._driver._hostutils.get_volume_info.assert_called_once_with(
            mock.sentinel.drive)
        mock_splitdrive.assert_called_once_with('fake_iscsi_lun_path')

    @mock.patch.object(windows_iscsi.WindowsISCSIDriver, '_get_capacity_info')
    def test_update_volume_stats(self, mock_get_capacity_info):
        mock_get_capacity_info.return_value = (
            mock.sentinel.size_gb,
            mock.sentinel.free_space_gb)

        self.flags(volume_backend_name='volume_backend_name')
        self.flags(reserved_percentage=10)

        expected_volume_stats = dict(
            volume_backend_name='volume_backend_name',
            vendor_name='Microsoft',
            driver_version=self._driver.VERSION,
            storage_protocol='iSCSI',
            total_capacity_gb=mock.sentinel.size_gb,
            free_capacity_gb=mock.sentinel.free_space_gb,
            reserved_percentage=10,
            QoS_support=False)

        self._driver._update_volume_stats()
        self.assertEqual(expected_volume_stats,
                         self._driver._stats)

    def test_extend_volume(self):
        volume = fake_volume.fake_volume_obj(mock.sentinel.fake_context)
        new_size_gb = 2
        expected_additional_sz_mb = 1024

        self._driver.extend_volume(volume, new_size_gb)

        self._driver._tgt_utils.extend_wt_disk.assert_called_once_with(
            volume.name, expected_additional_sz_mb)



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for volume transfer code.
"""

import mock
from oslo_serialization import jsonutils
from six.moves import http_client
import webob

from cinder.api.contrib import volume_transfer
from cinder import context
from cinder import db
from cinder import exception
from cinder.objects import fields
from cinder import quota
from cinder import test
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v2 import fakes as v2_fakes
from cinder.tests.unit import fake_constants as fake
import cinder.transfer


class VolumeTransferAPITestCase(test.TestCase):
    """Test Case for transfers API."""

    def setUp(self):
        super(VolumeTransferAPITestCase, self).setUp()
        self.volume_transfer_api = cinder.transfer.API()
        self.controller = volume_transfer.VolumeTransferController()
        self.user_ctxt = context.RequestContext(
            fake.USER_ID, fake.PROJECT_ID, auth_token=True, is_admin=True)

    def _create_transfer(self, volume_id=fake.VOLUME_ID,
                         display_name='test_transfer'):
        """Create a transfer object."""
        return self.volume_transfer_api.create(context.get_admin_context(),
                                               volume_id,
                                               display_name)

    @staticmethod
    def _create_volume(display_name='test_volume',
                       display_description='this is a test volume',
                       status='available',
                       size=1,
                       project_id=fake.PROJECT_ID,
                       attach_status=fields.VolumeAttachStatus.DETACHED):
        """Create a volume object."""
        vol = {}
        vol['host'] = 'fake_host'
        vol['size'] = size
        vol['user_id'] = fake.USER_ID
        vol['project_id'] = project_id
        vol['status'] = status
        vol['display_name'] = display_name
        vol['display_description'] = display_description
        vol['attach_status'] = attach_status
        vol['availability_zone'] = 'fake_zone'
        vol['volume_type_id'] = fake.VOLUME_TYPE_ID
        return db.volume_create(context.get_admin_context(), vol)['id']

    def test_show_transfer(self):
        volume_id = self._create_volume(size=5)
        transfer = self._create_transfer(volume_id)
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s' % (
            fake.PROJECT_ID, transfer['id']))
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)
        self.assertEqual(http_client.OK, res.status_int)
        self.assertEqual('test_transfer', res_dict['transfer']['name'])
        self.assertEqual(transfer['id'], res_dict['transfer']['id'])
        self.assertEqual(volume_id, res_dict['transfer']['volume_id'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_show_transfer_with_transfer_NotFound(self):
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s' % (
            fake.PROJECT_ID, fake.WILL_NOT_BE_FOUND_ID))
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.NOT_FOUND, res.status_int)
        self.assertEqual(http_client.NOT_FOUND,
                         res_dict['itemNotFound']['code'])
        self.assertEqual('Transfer %s could not be found.' %
                         fake.WILL_NOT_BE_FOUND_ID,
                         res_dict['itemNotFound']['message'])

    def test_list_transfers_json(self):
        volume_id_1 = self._create_volume(size=5)
        volume_id_2 = self._create_volume(size=5)
        transfer1 = self._create_transfer(volume_id_1)
        transfer2 = self._create_transfer(volume_id_2)

        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.OK, res.status_int)
        self.assertEqual(4, len(res_dict['transfers'][0]))
        self.assertEqual(transfer1['id'], res_dict['transfers'][0]['id'])
        self.assertEqual('test_transfer', res_dict['transfers'][0]['name'])
        self.assertEqual(4, len(res_dict['transfers'][1]))
        self.assertEqual('test_transfer', res_dict['transfers'][1]['name'])

        db.transfer_destroy(context.get_admin_context(), transfer2['id'])
        db.transfer_destroy(context.get_admin_context(), transfer1['id'])
        db.volume_destroy(context.get_admin_context(), volume_id_1)
        db.volume_destroy(context.get_admin_context(), volume_id_2)

    def test_list_transfers_detail_json(self):
        volume_id_1 = self._create_volume(size=5)
        volume_id_2 = self._create_volume(size=5)
        transfer1 = self._create_transfer(volume_id_1)
        transfer2 = self._create_transfer(volume_id_2)

        req = webob.Request.blank('/v2/%s/os-volume-transfer/detail' %
                                  fake.PROJECT_ID)
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        req.headers['Accept'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.OK, res.status_int)
        self.assertEqual(5, len(res_dict['transfers'][0]))
        self.assertEqual('test_transfer',
                         res_dict['transfers'][0]['name'])
        self.assertEqual(transfer1['id'], res_dict['transfers'][0]['id'])
        self.assertEqual(volume_id_1, res_dict['transfers'][0]['volume_id'])

        self.assertEqual(5, len(res_dict['transfers'][1]))
        self.assertEqual('test_transfer',
                         res_dict['transfers'][1]['name'])
        self.assertEqual(transfer2['id'], res_dict['transfers'][1]['id'])
        self.assertEqual(volume_id_2, res_dict['transfers'][1]['volume_id'])

        db.transfer_destroy(context.get_admin_context(), transfer2['id'])
        db.transfer_destroy(context.get_admin_context(), transfer1['id'])
        db.volume_destroy(context.get_admin_context(), volume_id_2)
        db.volume_destroy(context.get_admin_context(), volume_id_1)

    def test_list_transfers_with_all_tenants(self):
        volume_id_1 = self._create_volume(size=5)
        volume_id_2 = self._create_volume(size=5, project_id=fake.PROJECT_ID)
        transfer1 = self._create_transfer(volume_id_1)
        transfer2 = self._create_transfer(volume_id_2)

        req = fakes.HTTPRequest.blank('/v2/%s/os-volume-transfer?'
                                      'all_tenants=1' % fake.PROJECT_ID,
                                      use_admin_context=True)
        res_dict = self.controller.index(req)

        expected = [(transfer1['id'], 'test_transfer'),
                    (transfer2['id'], 'test_transfer')]
        ret = []
        for item in res_dict['transfers']:
            ret.append((item['id'], item['name']))
        self.assertEqual(set(expected), set(ret))

        db.transfer_destroy(context.get_admin_context(), transfer2['id'])
        db.transfer_destroy(context.get_admin_context(), transfer1['id'])
        db.volume_destroy(context.get_admin_context(), volume_id_1)

    def test_create_transfer_json(self):
        volume_id = self._create_volume(status='available', size=5)
        body = {"transfer": {"name": "transfer1",
                             "volume_id": volume_id}}

        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))

        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.ACCEPTED, res.status_int)
        self.assertIn('id', res_dict['transfer'])
        self.assertIn('auth_key', res_dict['transfer'])
        self.assertIn('created_at', res_dict['transfer'])
        self.assertIn('name', res_dict['transfer'])
        self.assertIn('volume_id', res_dict['transfer'])

        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_create_transfer_with_no_body(self):
        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.body = jsonutils.dump_as_bytes(None)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.headers['Accept'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])

    def test_create_transfer_with_body_KeyError(self):
        body = {"transfer": {"name": "transfer1"}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])

    def test_create_transfer_with_invalid_volume_id_value(self):
        body = {"transfer": {"name": "transfer1",
                             "volume_id": 1234}}

        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])

    def test_create_transfer_with_InvalidVolume(self):
        volume_id = self._create_volume(status='attached')
        body = {"transfer": {"name": "transfer1",
                             "volume_id": volume_id}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])
        self.assertEqual('Invalid volume: status must be available',
                         res_dict['badRequest']['message'])

        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_create_transfer_with_leading_trailing_spaces_for_name(self):
        volume_id = self._create_volume(status='available', size=5)
        body = {"transfer": {"name": "    transfer1   ",
                             "volume_id": volume_id}}

        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
                                  fake.PROJECT_ID)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))

        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.ACCEPTED, res.status_int)
        self.assertEqual(body['transfer']['name'].strip(),
                         res_dict['transfer']['name'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_delete_transfer_awaiting_transfer(self):
        volume_id = self._create_volume()
        transfer = self._create_transfer(volume_id)
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s' % (
                                  fake.PROJECT_ID, transfer['id']))
        req.method = 'DELETE'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))

        self.assertEqual(http_client.ACCEPTED, res.status_int)

        # verify transfer has been deleted
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s' % (
            fake.PROJECT_ID, transfer['id']))
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.NOT_FOUND, res.status_int)
        self.assertEqual(http_client.NOT_FOUND,
                         res_dict['itemNotFound']['code'])
        self.assertEqual('Transfer %s could not be found.' % transfer['id'],
                         res_dict['itemNotFound']['message'])
        self.assertEqual(db.volume_get(context.get_admin_context(),
                         volume_id)['status'], 'available')

        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_delete_transfer_with_transfer_NotFound(self):
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s' % (
            fake.PROJECT_ID, fake.WILL_NOT_BE_FOUND_ID))
        req.method = 'DELETE'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.NOT_FOUND, res.status_int)
        self.assertEqual(http_client.NOT_FOUND,
                         res_dict['itemNotFound']['code'])
        self.assertEqual('Transfer %s could not be found.' %
                         fake.WILL_NOT_BE_FOUND_ID,
                         res_dict['itemNotFound']['message'])

    @mock.patch.object(quota.QUOTAS, 'reserve')
    @mock.patch.object(db, 'volume_type_get', v2_fakes.fake_volume_type_get)
    def test_accept_transfer_volume_id_specified_json(self, type_get):
        volume_id = self._create_volume()
        transfer = self._create_transfer(volume_id)

        svc = self.start_service('volume', host='fake_host')
        body = {"accept": {"auth_key": transfer['auth_key']}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.ACCEPTED, res.status_int)
        self.assertEqual(transfer['id'], res_dict['transfer']['id'])
        self.assertEqual(volume_id, res_dict['transfer']['volume_id'])
        # cleanup
        svc.stop()

    def test_accept_transfer_with_no_body(self):
        volume_id = self._create_volume(size=5)
        transfer = self._create_transfer(volume_id)

        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))
        req.body = jsonutils.dump_as_bytes(None)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.headers['Accept'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_accept_transfer_with_body_KeyError(self):
        volume_id = self._create_volume(size=5)
        transfer = self._create_transfer(volume_id)

        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))
        body = {"": {}}
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.headers['Accept'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))

        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_accept_transfer_invalid_id_auth_key(self):
        volume_id = self._create_volume()
        transfer = self._create_transfer(volume_id)

        body = {"accept": {"auth_key": 1}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])
        self.assertEqual(res_dict['badRequest']['message'],
                         'Invalid auth key: Attempt to transfer %s with '
                         'invalid auth key.' % transfer['id'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_accept_transfer_with_invalid_transfer(self):
        volume_id = self._create_volume()
        transfer = self._create_transfer(volume_id)

        body = {"accept": {"auth_key": 1}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
            fake.PROJECT_ID, fake.WILL_NOT_BE_FOUND_ID))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.NOT_FOUND, res.status_int)
        self.assertEqual(http_client.NOT_FOUND,
                         res_dict['itemNotFound']['code'])
        self.assertEqual('Transfer %s could not be found.' %
                         fake.WILL_NOT_BE_FOUND_ID,
                         res_dict['itemNotFound']['message'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_accept_transfer_with_VolumeSizeExceedsAvailableQuota(self):

        def fake_transfer_api_accept_throwing_VolumeSizeExceedsAvailableQuota(
                cls, context, transfer, volume_id):
            raise exception.VolumeSizeExceedsAvailableQuota(requested='2',
                                                            consumed='2',
                                                            quota='3')

        self.mock_object(
            cinder.transfer.API,
            'accept',
            fake_transfer_api_accept_throwing_VolumeSizeExceedsAvailableQuota)

        volume_id = self._create_volume()
        transfer = self._create_transfer(volume_id)

        body = {"accept": {"auth_key": transfer['auth_key']}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))

        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(413, res.status_int)
        self.assertEqual(413, res_dict['overLimit']['code'])
        self.assertEqual('Requested volume or snapshot exceeds allowed '
                         'gigabytes quota. Requested 2G, quota is 3G and '
                         '2G has been consumed.',
                         res_dict['overLimit']['message'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_accept_transfer_with_VolumeLimitExceeded(self):

        def fake_transfer_api_accept_throwing_VolumeLimitExceeded(cls,
                                                                  context,
                                                                  transfer,
                                                                  volume_id):
            raise exception.VolumeLimitExceeded(allowed=1)

        self.mock_object(cinder.transfer.API, 'accept',
                         fake_transfer_api_accept_throwing_VolumeLimitExceeded)

        volume_id = self._create_volume()
        transfer = self._create_transfer(volume_id)

        body = {"accept": {"auth_key": transfer['auth_key']}}
        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))

        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(413, res.status_int)
        self.assertEqual(413, res_dict['overLimit']['code'])
        self.assertEqual("VolumeLimitExceeded: Maximum number of volumes "
                         "allowed (1) exceeded for quota 'volumes'.",
                         res_dict['overLimit']['message'])

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)

    def test_accept_transfer_with_auth_key_null(self):
        volume_id = self._create_volume(size=5)
        transfer = self._create_transfer(volume_id)
        body = {"accept": {"auth_key": None}}

        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                  fake.PROJECT_ID, transfer['id']))
        req.body = jsonutils.dump_as_bytes(body)
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.headers['Accept'] = 'application/json'
        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        res_dict = jsonutils.loads(res.body)

        self.assertEqual(http_client.BAD_REQUEST,
                         res_dict['badRequest']['code'])
        self.assertEqual(http_client.BAD_REQUEST, res.status_int)

        db.transfer_destroy(context.get_admin_context(), transfer['id'])
        db.volume_destroy(context.get_admin_context(), volume_id)
